from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, Union

import typing_extensions

from ...nodes import YulCase
from ..utils import IrInitTuple
from .abc import YulAbc
from .block import Block
from .literal import Literal

if TYPE_CHECKING:
    from .switch import Switch


class Case(YulAbc):
    """
    TBD
    """

    _parent: Switch
    _body: Block
    _value: Union[typing_extensions.Literal["default"], Literal]

    def __init__(self, init: IrInitTuple, case_: YulCase, parent: YulAbc):
        super().__init__(init, case_, parent)
        self._body = Block(init, case_.body, self)
        if case_.value == "default":
            self._value = "default"
        else:
            self._value = Literal(init, case_.value, self)

    def __iter__(self) -> Iterator[YulAbc]:
        yield self
        yield from self._body
        if self._value != "default":
            yield from self._value

    @property
    def parent(self) -> Switch:
        return self._parent

    @property
    def body(self) -> Block:
        return self._body

    @property
    def value(self) -> Union[typing_extensions.Literal["default"], Literal]:
        return self._value



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Union

from typing_extensions import Literal

if TYPE_CHECKING:
    from .core import Account, Chain, Wei
    from .transactions import TransactionAbc


class ChainBlocks:
    _chain: Chain
    _blocks: Dict[int, Block]

    def __init__(self, chain: Chain):
        self._chain = chain
        self._blocks = {}

    def __getitem__(
        self,
        key: Union[
            int,
            Literal["latest"],
            Literal["pending"],
            Literal["earliest"],
            Literal["safe"],
            Literal["finalized"],
        ],
    ) -> Block:
        if isinstance(key, int) and key < 0:
            key = self._chain.chain_interface.get_block_number() + key + 1
        if key not in self._blocks:
            data = self._chain.chain_interface.get_block(key)
            if data is None:
                raise KeyError(key)

            block = Block(self._chain, data)

            if "number" in data and data["number"] is not None:
                block_number = int(data["number"], 16)
                if block_number in self._blocks:
                    return self._blocks[block_number]

                if block.number <= self._chain.chain_interface.get_block_number():
                    self._blocks[block.number] = block
        else:
            block = self._blocks[key]
        return block

    def __len__(self):
        return self["latest"].number - self["earliest"].number + 1

    def __iter__(self) -> Iterator[Block]:
        for i in range(self["earliest"].number, self["latest"].number + 1):
            yield self[i]


class Block:
    _chain: Chain
    _block_data: Dict[str, Any]

    def __init__(self, chain: Chain, block_data: Dict[str, Any]):
        self._chain = chain
        self._block_data = block_data

    @property
    def chain(self) -> Chain:
        return self._chain

    @property
    def hash(self) -> str:
        return self._block_data["hash"]

    @property
    def parent_hash(self) -> str:
        return self._block_data["parentHash"]

    @property
    def sha3_uncles(self) -> str:
        return self._block_data["sha3Uncles"]

    @property
    def miner(self) -> Account:
        from .core import Account

        return Account(self._block_data["miner"], self._chain)

    @property
    def state_root(self) -> str:
        return self._block_data["stateRoot"]

    @property
    def transactions_root(self) -> str:
        return self._block_data["transactionsRoot"]

    @property
    def receipts_root(self) -> str:
        return self._block_data["receiptsRoot"]

    @property
    def number(self) -> int:
        return int(self._block_data["number"], 16)

    @property
    def gas_used(self) -> int:
        return int(self._block_data["gasUsed"], 16)

    @property
    def gas_limit(self) -> int:
        return int(self._block_data["gasLimit"], 16)

    @property
    def logs_bloom(self) -> str:
        return self._block_data["logsBloom"]

    @property
    def timestamp(self) -> int:
        return int(self._block_data["timestamp"], 16)

    @property
    def difficulty(self) -> int:
        return int(self._block_data["difficulty"], 16)

    @property
    def total_difficulty(self) -> int:
        return int(self._block_data["totalDifficulty"], 16)

    @property
    def seal_fields(self) -> Optional[List[str]]:
        if (
            "sealFields" in self._block_data
            and self._block_data["sealFields"] is not None
        ):
            return list(self._block_data["sealFields"])
        return None

    @property
    def uncles(self) -> List[str]:
        return list(self._block_data["uncles"])

    @property
    def txs(self) -> List[TransactionAbc]:
        return [self._chain.txs[tx] for tx in self._block_data["transactions"]]

    @property
    def size(self) -> int:
        return int(self._block_data["size"], 16)

    @property
    def mix_hash(self) -> str:
        return self._block_data["mixHash"]

    @property
    def nonce(self) -> str:
        return self._block_data["nonce"]

    @property
    def extra_data(self) -> str:
        return self._block_data["extraData"]

    @property
    def base_fee_per_gas(self) -> Optional[Wei]:
        if (
            "baseFeePerGas" in self._block_data
            and self._block_data["baseFeePerGas"] is not None
        ):
            from .core import Wei

            return Wei(int(self._block_data["baseFeePerGas"], 16))
        return None



import itertools
from pathlib import Path
from typing import AbstractSet

from woke.config import WokeConfig

from .exceptions import CompilationResolveError


class SourcePathResolver:
    __config: WokeConfig

    def __init__(self, woke_config: WokeConfig):
        self.__config = woke_config

    def resolve(
        self,
        source_unit_name: str,
        parent_source_unit_name: str,
        virtual_files: AbstractSet[Path],
    ) -> Path:
        """
        Return a system path for the given source unit name. Currently, this is done in a single step:
        - Try to find a source file in the Woke project directory or directories provided in the `include_paths` config option.
        There may be more steps implemented in the future. For example, we can add support for GitHub URLs.
        NPM packages can be easily resolved using the `include_paths` config option.
        """
        matching_paths = []

        for include_path in itertools.chain(
            [self.__config.project_root_path],
            self.__config.compiler.solc.include_paths,
            [Path(__file__).parent.parent / "contracts"],
        ):
            path = include_path / source_unit_name
            if path.is_file():
                matching_paths.append(path)
            elif path in virtual_files:
                matching_paths.append(path)

        if len(matching_paths) == 0:
            raise CompilationResolveError(
                f"Unable to find '{source_unit_name}' imported from '{parent_source_unit_name}' in the project root dir or include paths."
            )

        if len(matching_paths) > 1:
            err = f"Source unit name '{source_unit_name}' imported from '{parent_source_unit_name}' is ambiguous. It can be included as:"
            for matching_path in matching_paths:
                err += f"\n{matching_path}"
            raise CompilationResolveError(err)
        return matching_paths[0]

    def matches(self, source_unit_name: str, file: Path) -> bool:
        """
        Return True if the given source unit name matches the given file path.
        """
        for include_path in itertools.chain(
            [self.__config.project_root_path],
            self.__config.compiler.solc.include_paths,
            [Path(__file__).parent.parent / "contracts"],
        ):
            path = include_path / source_unit_name
            if path == file:
                return True
        return False



from typing import List, Optional, Set

from woke.analysis.detectors import DetectorAbc, DetectorResult, detector
from woke.analysis.detectors.utils import get_function_definition_from_expression
from woke.ast.enums import DataLocation, GlobalSymbolsEnum
from woke.ast.ir.declaration.variable_declaration import VariableDeclaration
from woke.ast.ir.expression.assignment import Assignment
from woke.ast.ir.expression.function_call import FunctionCall
from woke.ast.ir.expression.identifier import Identifier
from woke.ast.ir.expression.index_access import IndexAccess
from woke.ast.ir.expression.member_access import MemberAccess
from woke.ast.ir.statement.abc import StatementAbc
from woke.ast.ir.statement.expression_statement import ExpressionStatement
from woke.ast.ir.statement.variable_declaration_statement import (
    VariableDeclarationStatement,
)
from woke.core.solidity_version import SolidityVersionRange, SolidityVersionRanges


def check_bug_empty_byte_array_copy(fc: FunctionCall) -> Optional[DetectorResult]:
    versions = fc.version_ranges
    affected_versions = SolidityVersionRanges(
        [SolidityVersionRange(None, None, "0.7.14", False)]
    )
    if len(versions & affected_versions) == 0:
        return None

    # detects data.push()
    if (
        fc.function_called != GlobalSymbolsEnum.BYTES_PUSH
        or not isinstance(fc.expression, MemberAccess)
        or not isinstance(fc.expression.expression, Identifier)
        or len(fc.arguments) != 0
    ):
        return None

    var_ident = fc.expression.expression
    if not var_ident.is_ref_to_state_variable or not var_ident:
        return None

    fn_def = get_function_definition_from_expression(fc)
    if not fn_def or not fn_def.body:
        return None

    var_ident_stmt = var_ident
    while var_ident_stmt.parent and not isinstance(var_ident_stmt, StatementAbc):
        var_ident_stmt = var_ident_stmt.parent
    if not var_ident_stmt:
        return None

    start = False
    memory_write = False
    data_decl = None
    local_variable_decl = None
    for stmt in reversed(list(fn_def.body.statements_iter())):
        if stmt == var_ident_stmt:
            start = True
        if not start:
            continue

        if isinstance(stmt, ExpressionStatement) and isinstance(
            stmt.expression, Assignment
        ):
            # detects data = t;
            if data_decl is None:
                if (
                    isinstance(stmt.expression.left_expression, Identifier)
                    and stmt.expression.left_expression.referenced_declaration
                    == var_ident.referenced_declaration
                    and stmt.expression.right_expression
                    and isinstance(stmt.expression.right_expression, Identifier)
                ):
                    data_decl = stmt.expression.left_expression.referenced_declaration
                    local_variable_decl = (
                        stmt.expression.right_expression.referenced_declaration
                    )
            # detects storing something in the memory
            elif (
                data_decl
                and isinstance(stmt.expression.left_expression, IndexAccess)
                and isinstance(
                    stmt.expression.left_expression.base_expression, Identifier
                )
                and isinstance(
                    stmt.expression.left_expression.base_expression.referenced_declaration,
                    VariableDeclaration,
                )
                and stmt.expression.left_expression.base_expression.referenced_declaration.data_location
                == DataLocation.MEMORY
            ):
                memory_write = True
        # detects variable declaration in memory
        elif (
            data_decl
            and memory_write
            and isinstance(stmt, VariableDeclarationStatement)
        ):
            for dec in stmt.declarations:
                if dec == local_variable_decl:
                    return DetectorResult(
                        fc, "Possible sequence leading to the empty byte array copy bug"
                    )
    return None


@detector(-1020, "bug-empty-byte-array-copy")
class BugEmptyByteArrayCopyDetector(DetectorAbc):
    """
    Detects empty array copy bug for solidity versions < 0.7.14
    (https://blog.soliditylang.org/2020/10/19/empty-byte-array-copy-bug/)
    """

    _detections: Set[DetectorResult]

    def __init__(self):
        self._detections = set()

    def report(self) -> List[DetectorResult]:
        return list(self._detections)

    def visit_function_call(self, node: FunctionCall):
        res = check_bug_empty_byte_array_copy(node)
        if res:
            self._detections.add(res)



from woke.config import WokeConfig
from woke.development.globals import reset_exception_handled


class PytestWokePlugin:
    _config: WokeConfig

    def __init__(self, config: WokeConfig):
        self._config = config

    def pytest_runtest_setup(self, item):
        reset_exception_handled()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Tuple

from .abc import DeclarationAbc

if TYPE_CHECKING:
    from .enum_definition import EnumDefinition

from woke.ast.ir.abc import SolidityAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import SolcEnumValue

logger = logging.getLogger(__name__)


class EnumValue(DeclarationAbc):
    """
    Definition of an enum value inside an enum definition.

    !!! example
        `GoLeft`, `GoRight`, `GoStraight`, `SitStill` in the following enum definition:
        ```solidity
        enum ActionChoices { GoLeft, GoRight, GoStraight, SitStill }
        ```
    """

    _ast_node: SolcEnumValue
    _parent: EnumDefinition

    def __init__(self, init: IrInitTuple, value: SolcEnumValue, parent: SolidityAbc):
        super().__init__(init, value, parent)

    def _parse_name_location(self) -> Tuple[int, int]:
        src = self._ast_node.src
        return src.byte_offset, src.byte_offset + src.byte_length

    @property
    def parent(self) -> EnumDefinition:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def canonical_name(self) -> str:
        return f"{self._parent.canonical_name}.{self._name}"

    @property
    def declaration_string(self) -> str:
        return self.name



from __future__ import annotations

import subprocess
import time
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union
from urllib.error import URLError

from typing_extensions import Literal, TypedDict

from woke.config import WokeConfig
from woke.utils.networking import get_free_port

from .json_rpc.communicator import JsonRpcCommunicator

TxParams = TypedDict(
    "TxParams",
    {
        "type": int,
        "nonce": int,
        "to": str,
        "from": str,
        "gas": Union[int, Literal["auto"]],
        "value": int,
        "data": bytes,
        "gasPrice": int,
        "maxPriorityFeePerGas": int,
        "maxFeePerGas": int,
        "accessList": Union[List, Literal["auto"]],
        "chainId": int,
    },
    total=False,
)


class ChainInterfaceAbc(ABC):
    _config: WokeConfig
    _communicator: JsonRpcCommunicator
    _process: Optional[subprocess.Popen]

    def __init__(
        self,
        config: WokeConfig,
        communicator: JsonRpcCommunicator,
        process: Optional[subprocess.Popen] = None,
    ) -> None:
        self._config = config
        self._communicator = communicator
        self._process = process

    @staticmethod
    def _encode_tx_params(transaction: TxParams) -> Dict:
        tx = {}
        if "type" in transaction:
            tx["type"] = hex(transaction["type"])
        if "nonce" in transaction:
            tx["nonce"] = hex(transaction["nonce"])
        if "to" in transaction:
            tx["to"] = transaction["to"]
        if "from" in transaction:
            tx["from"] = transaction["from"]
        if "gas" in transaction:
            assert transaction["gas"] != "auto"
            tx["gas"] = hex(transaction["gas"])
        if "value" in transaction:
            tx["value"] = hex(transaction["value"])
        if "data" in transaction:
            tx["data"] = "0x" + transaction["data"].hex()
        if "gasPrice" in transaction:
            tx["gasPrice"] = hex(transaction["gasPrice"])
        if "maxPriorityFeePerGas" in transaction:
            tx["maxPriorityFeePerGas"] = hex(transaction["maxPriorityFeePerGas"])
        if "maxFeePerGas" in transaction:
            tx["maxFeePerGas"] = hex(transaction["maxFeePerGas"])
        if "accessList" in transaction:
            assert transaction["accessList"] != "auto"
            tx["accessList"] = transaction["accessList"]
        if "chainId" in transaction:
            tx["chainId"] = hex(transaction["chainId"])
        return tx

    @staticmethod
    def _encode_block_identifier(block_identifier: Union[int, str]) -> str:
        if isinstance(block_identifier, int):
            return hex(block_identifier)
        elif isinstance(block_identifier, str):
            return block_identifier
        else:
            raise TypeError("block identifier must be either int or str")

    @classmethod
    def launch(
        cls,
        config: WokeConfig,
        *,
        accounts: Optional[int] = None,
        chain_id: Optional[int] = None,
        fork: Optional[str] = None,
        hardfork: Optional[str] = None,
    ) -> ChainInterfaceAbc:
        if config.testing.cmd == "anvil":
            args = ["anvil"] + config.testing.anvil.cmd_args.split()
            constructor = AnvilChainInterface
        elif config.testing.cmd == "ganache":
            args = ["ganache"] + config.testing.ganache.cmd_args.split()
            constructor = GanacheChainInterface
        elif config.testing.cmd == "hardhat":
            if (
                accounts is not None
                or chain_id is not None
                or fork is not None
                or hardfork is not None
            ):
                raise ValueError(
                    "Setting accounts, chain_id, fork and hardfork is not supported for hardhat"
                )

            args = ["npx", "hardhat", "node"] + config.testing.hardhat.cmd_args.split()
            constructor = HardhatChainInterface
        else:
            raise NotImplementedError(f"Network {config.testing.cmd} not supported")

        hostname = "127.0.0.1"
        port = None
        accounts_set = False
        chain_id_set = False
        fork_set = False
        hardfork_set = False

        for i, arg in enumerate(args):
            if arg in {"--port", "-p", "--server.port"}:
                try:
                    port = args[i + 1]
                except IndexError:
                    port = "8545"
            elif arg in {"--hostname", "-h", "--host", "--server.hostname"}:
                try:
                    hostname = args[i + 1]
                except IndexError:
                    hostname = "127.0.0.1"
            elif (
                arg in {"-a", "--accounts", "--wallet.accounts"}
                and accounts is not None
            ):
                accounts_set = True
                try:
                    args[i + 1] = str(accounts)
                except IndexError:
                    args += [str(accounts)]
            elif arg in {"--chain-id", "--chain.chainId"} and chain_id is not None:
                chain_id_set = True
                try:
                    args[i + 1] = str(chain_id)
                except IndexError:
                    args += [str(chain_id)]
            elif (
                arg in {"-f", "--fork-url", "--fork.url", "--rpc-url"}
                and fork is not None
            ):
                fork_set = True
                try:
                    args[i + 1] = fork
                except IndexError:
                    args += [fork]
            elif (
                arg in {"--hardfork", "-k", "--chain.hardfork"} and hardfork is not None
            ):
                hardfork_set = True
                try:
                    args[i + 1] = hardfork
                except IndexError:
                    args += [hardfork]

        if port is None:
            port = str(get_free_port())
            args += ["--port", port]
        if accounts is not None and not accounts_set:
            args += ["-a", str(accounts)]
        if chain_id is not None and not chain_id_set:
            if config.testing.cmd == "anvil":
                args += ["--chain-id", str(chain_id)]
            elif config.testing.cmd == "ganache":
                args += ["--chain.chainId", str(chain_id)]
        if fork is not None and not fork_set:
            args += ["-f", fork]
        if hardfork is not None and not hardfork_set:
            if config.testing.cmd == "anvil":
                args += ["--hardfork", hardfork]
            elif config.testing.cmd == "ganache":
                args += ["-k", hardfork]

        print(f"Launching {' '.join(args)}")
        process = subprocess.Popen(args, stdout=subprocess.DEVNULL)

        try:
            start = time.perf_counter()
            while True:
                try:
                    comm = JsonRpcCommunicator(config, f"ws://{hostname}:{port}")
                    comm.__enter__()
                    comm.send_request("web3_clientVersion").lower()
                    break
                except (ConnectionRefusedError, URLError, ValueError):
                    if time.perf_counter() - start > config.testing.timeout:
                        raise
                    time.sleep(0.05)

            return constructor(config, comm, process)
        except Exception:
            if process.returncode is None:
                process.terminate()
                try:
                    process.communicate(timeout=5)
                except subprocess.TimeoutExpired:
                    process.kill()
            raise

    @classmethod
    def connect(cls, config: WokeConfig, uri: str) -> ChainInterfaceAbc:
        communicator = JsonRpcCommunicator(config, uri)
        communicator.__enter__()
        try:
            client_version: str = communicator.send_request(
                "web3_clientVersion"
            ).lower()
            chain_id = int(communicator.send_request("eth_chainId"), 16)
            if "anvil" in client_version:
                return AnvilChainInterface(config, communicator)
            elif "hardhat" in client_version:
                return HardhatChainInterface(config, communicator)
            elif "ethereumjs" in client_version:
                return GanacheChainInterface(config, communicator)
            elif client_version.startswith(("geth", "bor")):
                return GethChainInterface(config, communicator)
            elif "nitro" in client_version:
                return NitroChainInterface(config, communicator)
            elif chain_id in {43113, 43114}:
                # Avax client reports just a version number without the name of the client
                # => hard to distinguish from other clients
                return AvalancheChainInterface(config, communicator)
            elif chain_id in {1101, 1442}:
                return HermezChainInterface(config, communicator)
            else:
                raise NotImplementedError(
                    f"Client version {client_version} not supported"
                )
        except Exception:
            communicator.__exit__(None, None, None)
            raise

    def close(self) -> None:
        self._communicator.__exit__(None, None, None)
        if self._process is not None:
            self._process.terminate()
            try:
                self._process.communicate(timeout=self._config.testing.timeout)
            except subprocess.TimeoutExpired:
                self._process.kill()

    def get_client_version(self) -> str:
        return self._communicator.send_request("web3_clientVersion")

    def get_balance(
        self, address: str, block_identifier: Union[int, str] = "latest"
    ) -> int:
        return int(
            self._communicator.send_request(
                "eth_getBalance",
                [address, self._encode_block_identifier(block_identifier)],
            ),
            16,
        )

    def get_code(
        self, address: str, block_identifier: Union[int, str] = "latest"
    ) -> bytes:
        return bytes.fromhex(
            self._communicator.send_request(
                "eth_getCode",
                [address, self._encode_block_identifier(block_identifier)],
            )[2:]
        )

    def get_coinbase(self) -> str:
        return self._communicator.send_request("eth_coinbase")

    def get_block(
        self, block_identifier: Union[int, str], include_transactions: bool = False
    ) -> Dict[str, Any]:
        return self._communicator.send_request(
            "eth_getBlockByNumber",
            [self._encode_block_identifier(block_identifier), include_transactions],
        )

    def get_block_number(self) -> int:
        return int(self._communicator.send_request("eth_blockNumber"), 16)

    def get_transaction(self, tx_hash: str) -> Dict[str, Any]:
        return self._communicator.send_request("eth_getTransactionByHash", [tx_hash])

    def get_chain_id(self) -> int:
        return int(self._communicator.send_request("eth_chainId"), 16)

    def get_gas_price(self) -> int:
        return int(self._communicator.send_request("eth_gasPrice"), 16)

    def get_transaction_count(
        self, address: str, block_identifier: Union[int, str] = "latest"
    ) -> int:
        return int(
            self._communicator.send_request(
                "eth_getTransactionCount",
                [address, self._encode_block_identifier(block_identifier)],
            ),
            16,
        )

    def get_transaction_receipt(self, tx_hash: str) -> Optional[Dict[str, Any]]:
        return self._communicator.send_request("eth_getTransactionReceipt", [tx_hash])

    def call(
        self, params: TxParams, block_identifier: Union[int, str] = "latest"
    ) -> bytes:
        return bytes.fromhex(
            self._communicator.send_request(
                "eth_call",
                [
                    self._encode_tx_params(params),
                    self._encode_block_identifier(block_identifier),
                ],
            )[2:]
        )

    def estimate_gas(
        self, params: TxParams, block_identifier: Union[int, str] = "pending"
    ) -> int:
        return int(
            self._communicator.send_request(
                "eth_estimateGas",
                [
                    self._encode_tx_params(params),
                    self._encode_block_identifier(block_identifier),
                ],
            ),
            16,
        )

    def send_transaction(self, params: TxParams) -> str:
        return self._communicator.send_request(
            "eth_sendTransaction", [self._encode_tx_params(params)]
        )

    def send_raw_transaction(self, raw_tx: bytes) -> str:
        return self._communicator.send_request(
            "eth_sendRawTransaction", ["0x" + raw_tx.hex()]
        )

    def debug_trace_transaction(
        self, tx_hash: str, options: Optional[Dict] = None
    ) -> Dict[str, Any]:
        return self._communicator.send_request(
            "debug_traceTransaction",
            [tx_hash, options] if options is not None else [tx_hash],
        )

    def debug_trace_call(
        self,
        params: TxParams,
        block_identifier: Union[int, str] = "latest",
        options: Optional[Dict] = None,
    ) -> Dict[str, Any]:
        return self._communicator.send_request(
            "debug_traceCall",
            [
                self._encode_tx_params(params),
                self._encode_block_identifier(block_identifier),
                options,
            ]
            if options is not None
            else [
                self._encode_tx_params(params),
                self._encode_block_identifier(block_identifier),
            ],
        )

    def snapshot(self) -> str:
        return self._communicator.send_request("evm_snapshot")

    def revert(self, snapshot_id: str) -> bool:
        return self._communicator.send_request("evm_revert", [snapshot_id])

    def mine(self, timestamp: Optional[int]) -> None:
        self._communicator.send_request(
            "evm_mine", [hex(timestamp)] if timestamp is not None else None
        )

    def get_max_priority_fee_per_gas(self) -> int:
        return int(self._communicator.send_request("eth_maxPriorityFeePerGas"), 16)

    def sign(self, address: str, message: bytes) -> bytes:
        return bytes.fromhex(
            self._communicator.send_request(
                "eth_sign", [address, "0x" + message.hex()]
            )[2:]
        )

    def sign_typed(self, address: str, message: Dict) -> bytes:
        return bytes.fromhex(
            self._communicator.send_request("eth_signTypedData_v4", [address, message])[
                2:
            ]
        )

    def create_access_list(
        self, params: TxParams, block_identifier: Union[int, str] = "pending"
    ) -> Dict[str, Any]:
        return self._communicator.send_request(
            "eth_createAccessList",
            [
                self._encode_tx_params(params),
                self._encode_block_identifier(block_identifier),
            ],
        )

    def trace_transaction(self, tx_hash: str) -> List:
        return self._communicator.send_request("trace_transaction", [tx_hash])

    def get_storage_at(
        self, address: str, position: int, block_identifier: Union[int, str] = "latest"
    ) -> bytes:
        return bytes.fromhex(
            self._communicator.send_request(
                "eth_getStorageAt",
                [
                    address,
                    hex(position),
                    self._encode_block_identifier(block_identifier),
                ],
            )[2:]
        )

    @abstractmethod
    def get_accounts(self) -> List[str]:
        ...

    @abstractmethod
    def get_automine(self) -> bool:
        ...

    @abstractmethod
    def set_automine(self, value: bool) -> None:
        ...

    @abstractmethod
    def reset(self, options: Optional[Dict] = None) -> None:
        ...

    @abstractmethod
    def set_coinbase(self, address: str) -> None:
        ...

    @abstractmethod
    def set_balance(self, address: str, value: int) -> None:
        ...

    @abstractmethod
    def set_block_gas_limit(self, gas_limit: int) -> None:
        ...

    @abstractmethod
    def set_code(self, address: str, value: bytes) -> None:
        ...

    @abstractmethod
    def set_nonce(self, address: str, value: int) -> None:
        ...

    @abstractmethod
    def set_next_block_timestamp(self, timestamp: int) -> None:
        ...

    @abstractmethod
    def send_unsigned_transaction(self, params: TxParams) -> str:
        ...

    @abstractmethod
    def set_next_block_base_fee_per_gas(self, value: int) -> None:
        ...

    @abstractmethod
    def set_min_gas_price(self, value: int) -> None:
        ...

    @abstractmethod
    def set_storage_at(self, address: str, position: int, value: bytes) -> None:
        ...

    @abstractmethod
    def mine_many(self, num_blocks: int, timestamp_change: Optional[int]) -> None:
        ...


class HardhatChainInterface(ChainInterfaceAbc):
    def get_accounts(self) -> List[str]:
        return self._communicator.send_request("eth_accounts")

    def set_balance(self, address: str, value: int) -> None:
        self._communicator.send_request("hardhat_setBalance", [address, hex(value)])

    def impersonate_account(self, address: str) -> None:
        self._communicator.send_request("hardhat_impersonateAccount", [address])

    def stop_impersonating_account(self, address: str) -> None:
        self._communicator.send_request("hardhat_stopImpersonatingAccount", [address])

    def set_block_gas_limit(self, gas_limit: int) -> None:
        self._communicator.send_request("evm_setBlockGasLimit", [hex(gas_limit)])

    def reset(self, options: Optional[Dict] = None) -> None:
        self._communicator.send_request(
            "hardhat_reset", [options] if options is not None else None
        )

    def set_coinbase(self, address: str) -> None:
        self._communicator.send_request("hardhat_setCoinbase", [address])

    def get_automine(self) -> bool:
        return self._communicator.send_request("hardhat_getAutomine")

    def set_automine(self, value: bool) -> None:
        self._communicator.send_request("evm_setAutomine", [value])

    def set_code(self, address: str, value: bytes) -> None:
        self._communicator.send_request(
            "hardhat_setCode", [address, "0x" + value.hex()]
        )

    def set_nonce(self, address: str, value: int) -> None:
        self._communicator.send_request("hardhat_setNonce", [address, hex(value)])

    def set_next_block_timestamp(self, timestamp: int) -> None:
        self._communicator.send_request("evm_setNextBlockTimestamp", [hex(timestamp)])

    def send_unsigned_transaction(self, params: TxParams) -> str:
        raise NotImplementedError(
            "Hardhat does not support sending unsigned transactions"
        )

    def set_next_block_base_fee_per_gas(self, value: int) -> None:
        self._communicator.send_request(
            "hardhat_setNextBlockBaseFeePerGas", [hex(value)]
        )

    def set_min_gas_price(self, value: int) -> None:
        self._communicator.send_request("hardhat_setMinGasPrice", [hex(value)])

    def set_storage_at(self, address: str, position: int, value: bytes) -> None:
        self._communicator.send_request(
            "hardhat_setStorageAt", [address, hex(position), "0x" + value.hex()]
        )

    def mine_many(self, num_blocks: int, timestamp_change: Optional[int]) -> None:
        self._communicator.send_request(
            "hardhat_mine",
            [hex(num_blocks), hex(timestamp_change)]
            if timestamp_change is not None
            else [hex(num_blocks)],
        )


class AnvilChainInterface(ChainInterfaceAbc):
    def get_accounts(self) -> List[str]:
        return self._communicator.send_request("eth_accounts")

    def set_balance(self, address: str, value: int) -> None:
        self._communicator.send_request("anvil_setBalance", [address, hex(value)])

    def impersonate_account(self, address: str) -> None:
        self._communicator.send_request("anvil_impersonateAccount", [address])

    def stop_impersonating_account(self, address: str) -> None:
        self._communicator.send_request("anvil_stopImpersonatingAccount", [address])

    def set_block_gas_limit(self, gas_limit: int) -> None:
        self._communicator.send_request("evm_setBlockGasLimit", [hex(gas_limit)])

    def reset(self, options: Optional[Dict] = None) -> None:
        self._communicator.send_request(
            "anvil_reset", [options] if options is not None else None
        )

    def set_coinbase(self, address: str) -> None:
        self._communicator.send_request("anvil_setCoinbase", [address])

    def get_automine(self) -> bool:
        return self._communicator.send_request("anvil_getAutomine")

    def set_automine(self, value: bool) -> None:
        self._communicator.send_request("evm_setAutomine", [value])

    def set_code(self, address: str, value: bytes) -> None:
        self._communicator.send_request("anvil_setCode", [address, "0x" + value.hex()])

    def set_nonce(self, address: str, value: int) -> None:
        self._communicator.send_request("anvil_setNonce", [address, hex(value)])

    def set_next_block_timestamp(self, timestamp: int) -> None:
        self._communicator.send_request("evm_setNextBlockTimestamp", [hex(timestamp)])

    def send_unsigned_transaction(self, params: TxParams) -> str:
        return self._communicator.send_request(
            "eth_sendUnsignedTransaction", [self._encode_tx_params(params)]
        )

    def set_next_block_base_fee_per_gas(self, value: int) -> None:
        self._communicator.send_request("anvil_setNextBlockBaseFeePerGas", [hex(value)])

    def set_min_gas_price(self, value: int) -> None:
        self._communicator.send_request("anvil_setMinGasPrice", [hex(value)])

    def set_storage_at(self, address: str, position: int, value: bytes) -> None:
        self._communicator.send_request(
            "anvil_setStorageAt", [address, hex(position), "0x" + value.hex()]
        )

    def node_info(self) -> Dict[str, Any]:
        return self._communicator.send_request("anvil_nodeInfo")

    def mine_many(self, num_blocks: int, timestamp_change: Optional[int]) -> None:
        self._communicator.send_request(
            "anvil_mine",
            [hex(num_blocks), hex(timestamp_change)]
            if timestamp_change is not None
            else [hex(num_blocks)],
        )


class GanacheChainInterface(ChainInterfaceAbc):
    def get_accounts(self) -> List[str]:
        return self._communicator.send_request("eth_accounts")

    def set_balance(self, address: str, value: int) -> None:
        self._communicator.send_request("evm_setAccountBalance", [address, hex(value)])

    def add_account(self, address: str, passphrase: str) -> bool:
        return self._communicator.send_request(
            "evm_addAccount", [address, passphrase]
        ) and self._communicator.send_request(
            "personal_unlockAccount", [address, passphrase, hex(0)]
        )

    def set_block_gas_limit(self, gas_limit: int) -> None:
        raise NotImplementedError("Ganache does not support setting block gas limit")

    def reset(self, options: Optional[Dict] = None) -> None:
        raise NotImplementedError("Ganache does not support resetting the chain")

    def set_coinbase(self, address: str) -> None:
        raise NotImplementedError("Ganache does not support setting coinbase")

    def get_automine(self) -> bool:
        raise NotImplementedError("Ganache does not support automine")

    def set_automine(self, value: bool) -> None:
        raise NotImplementedError("Ganache does not support automine")

    def set_code(self, address: str, value: bytes) -> None:
        self._communicator.send_request(
            "evm_setAccountCode", [address, "0x" + value.hex()]
        )

    def set_nonce(self, address: str, value: int) -> None:
        self._communicator.send_request("evm_setAccountNonce", [address, hex(value)])

    def set_next_block_timestamp(self, timestamp: int) -> None:
        raise NotImplementedError(
            "Ganache does not support setting next block timestamp"
        )

    def send_unsigned_transaction(self, params: TxParams) -> str:
        raise NotImplementedError(
            "Ganache does not support sending unsigned transactions"
        )

    def set_next_block_base_fee_per_gas(self, value: int) -> None:
        raise NotImplementedError(
            "Ganache does not support setting next block base fee per gas"
        )

    def set_min_gas_price(self, value: int) -> None:
        self._communicator.send_request("miner_setGasPrice", [hex(value)])

    def set_storage_at(self, address: str, position: int, value: bytes) -> None:
        self._communicator.send_request(
            "evm_setAccountStorageAt", [address, hex(position), "0x" + value.hex()]
        )

    def mine_many(self, num_blocks: int, timestamp_change: Optional[int]) -> None:
        if timestamp_change is not None:
            raise NotImplementedError(
                "Ganache does not support timestamp intervals when mining multiple blocks"
            )
        self._communicator.send_request(
            "evm_mine",
            [{"blocks": num_blocks}],
        )


class GethLikeChainInterfaceAbc(ChainInterfaceAbc, ABC):
    @property
    @abstractmethod
    def _name(self) -> str:
        ...

    def get_accounts(self) -> List[str]:
        return []

    def get_automine(self) -> bool:
        raise NotImplementedError(f"{self._name} does not support automine")

    def set_automine(self, value: bool) -> None:
        raise NotImplementedError(f"{self._name} does not support automine")

    def reset(self, options: Optional[Dict] = None) -> None:
        raise NotImplementedError(f"{self._name} does not support resetting the chain")

    def set_coinbase(self, address: str) -> None:
        raise NotImplementedError(f"{self._name} does not support setting coinbase")

    def set_balance(self, address: str, value: int) -> None:
        raise NotImplementedError(f"{self._name} does not support setting balance")

    def set_block_gas_limit(self, gas_limit: int) -> None:
        raise NotImplementedError(
            f"{self._name} does not support setting block gas limit"
        )

    def set_code(self, address: str, value: bytes) -> None:
        raise NotImplementedError(f"{self._name} does not support setting code")

    def set_nonce(self, address: str, value: int) -> None:
        raise NotImplementedError(f"{self._name} does not support setting nonce")

    def set_next_block_timestamp(self, timestamp: int) -> None:
        raise NotImplementedError(
            f"{self._name} does not support setting next block timestamp"
        )

    def send_unsigned_transaction(self, params: TxParams) -> str:
        raise NotImplementedError(
            f"{self._name} does not support sending unsigned transactions"
        )

    def set_next_block_base_fee_per_gas(self, value: int) -> None:
        raise NotImplementedError(
            f"{self._name} does not support setting next block base fee per gas"
        )

    def set_min_gas_price(self, value: int) -> None:
        raise NotImplementedError(
            f"{self._name} does not support setting min gas price"
        )

    def set_storage_at(self, address: str, position: int, value: bytes) -> None:
        raise NotImplementedError(f"{self._name} does not support setting storage")

    def mine_many(self, num_blocks: int, timestamp_change: Optional[int]) -> None:
        raise NotImplementedError(f"{self._name} does not support mining blocks")


class GethChainInterface(GethLikeChainInterfaceAbc):
    @property
    def _name(self) -> str:
        return "Geth"

    def get_accounts(self) -> List[str]:
        return self._communicator.send_request("eth_accounts")


class HermezChainInterface(GethLikeChainInterfaceAbc):
    @property
    def _name(self) -> str:
        return "Hermez"


class NitroChainInterface(GethLikeChainInterfaceAbc):
    @property
    def _name(self) -> str:
        return "Nitro"


class AvalancheChainInterface(GethLikeChainInterfaceAbc):
    @property
    def _name(self) -> str:
        return "Avalanche"



import logging
from typing import List

import graphviz as gv

from woke.ast.ir.declaration.function_definition import FunctionDefinition
from woke.ast.ir.declaration.modifier_definition import ModifierDefinition
from woke.ast.ir.statement.abc import StatementAbc
from woke.lsp.common_structures import DocumentUri
from woke.lsp.context import LspContext
from woke.lsp.exceptions import LspError
from woke.lsp.protocol_structures import ErrorCodes
from woke.lsp.utils import uri_to_path

logger = logging.getLogger(__name__)


async def generate_cfg_handler(
    context: LspContext, uri: DocumentUri, canonical_name: str
) -> str:
    logger.debug(
        f"Control flow graph for function {canonical_name} in file {uri} requested"
    )
    await context.compiler.output_ready.wait()

    path = uri_to_path(uri).resolve()

    if path not in context.compiler.source_units:
        raise LspError(ErrorCodes.InternalError, "File not found in compiler output")

    source_unit = context.compiler.source_units[path]
    target_declaration = None
    for declaration in source_unit.declarations_iter():
        if declaration.canonical_name == canonical_name:
            if not isinstance(declaration, (FunctionDefinition, ModifierDefinition)):
                raise LspError(
                    ErrorCodes.InvalidParams,
                    "Declaration is not a function or modifier",
                )
            target_declaration = declaration
            break

    if target_declaration is None:
        raise LspError(ErrorCodes.InvalidParams, "Declaration not found")

    cfg = target_declaration.cfg
    if cfg is None:
        raise LspError(ErrorCodes.InternalError, "Control flow graph not available")
    graph = cfg.graph

    g = gv.Digraph(f"{canonical_name} control flow graph")
    g.attr(rankdir=context.config.generator.control_flow_graph.direction)
    g.attr("node", shape="box")

    skip_start_block = False
    if (
        len(cfg.start_block.statements) == 0
        and cfg.start_block.control_statement is None
        and graph.out_degree(cfg.start_block)  # pyright: ignore reportGeneralTypeIssues
        == 1
    ):
        skip_start_block = True

    skip_end_block = False
    if (
        len(cfg.end_block.statements) == 0
        and cfg.end_block.control_statement is None
        and graph.in_degree(cfg.end_block)  # pyright: ignore reportGeneralTypeIssues
        == 1
    ):
        skip_end_block = True

    for node in graph.nodes:  # pyright: ignore reportGeneralTypeIssues
        if skip_start_block and node == cfg.start_block:
            continue
        if skip_end_block and node == cfg.end_block:
            continue

        statements: List[StatementAbc] = node.statements
        node_attrs = {
            "label": "".join(
                f"{line}\l"  # pyright: ignore reportInvalidStringEscapeSequence
                for line in str(node).splitlines()
            )
        }
        if (
            context.config.generator.control_flow_graph.vscode_urls
            and len(statements) > 0
        ):
            first_statement = statements[0]
            line, column = context.compiler.get_line_pos_from_byte_offset(
                first_statement.file, first_statement.byte_location[0]
            )
            line += 1
            column += 1
            node_attrs["URL"] = f"vscode://file/{first_statement.file}:{line}:{column}"
        g.node(str(node.id), **node_attrs)

    for (
        from_,
        to,
        data,
    ) in graph.edges.data():  # pyright: ignore reportGeneralTypeIssues
        if skip_start_block and from_ == cfg.start_block:
            continue
        if skip_end_block and to == cfg.end_block:
            continue

        condition = data["condition"]  # pyright: ignore reportOptionalSubscript
        if condition[1] is not None:
            label = f"{condition[1].source} {condition[0]}"
        else:
            label = condition[0]
        g.edge(str(from_.id), str(to.id), label=label)

    return g.source



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, List, Optional, Union

from ...enums import ModifierInvocationKind
from ..expression.abc import ExpressionAbc
from ..expression.identifier import Identifier
from ..reference_resolver import CallbackParams
from .identifier_path import IdentifierPath

if TYPE_CHECKING:
    from ..declaration.function_definition import FunctionDefinition

from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import SolcIdentifier, SolcIdentifierPath, SolcModifierInvocation


class ModifierInvocation(SolidityAbc):
    """
    !!! warning
        Also represents a base constructor invocation.
    !!! example
        - `:::solidity IERC20("MyToken", "MTK")` ([ModifierInvocationKind.BASE_CONSTRUCTOR_SPECIFIER][woke.ast.enums.ModifierInvocationKind.BASE_CONSTRUCTOR_SPECIFIER]),
        - `:::solidity initializer` ([ModifierInvocationKind.MODIFIER_INVOCATION][woke.ast.enums.ModifierInvocationKind.MODIFIER_INVOCATION])

        in the following code:
        ```solidity
        import Initializable from "@openzeppelin/contracts/proxy/utils/Initializable.sol";

        contract MyContract is ERC20, Initializable {
            constructor() ERC20("MyToken", "MTK") initializer {
                // ...
            }
        }
        ```
    """

    _ast_node: SolcModifierInvocation
    _parent: FunctionDefinition

    _kind: Optional[ModifierInvocationKind]
    _modifier_name: Union[Identifier, IdentifierPath]
    _arguments: Optional[List[ExpressionAbc]]

    def __init__(
        self,
        init: IrInitTuple,
        modifier_invocation: SolcModifierInvocation,
        parent: SolidityAbc,
    ):
        super().__init__(init, modifier_invocation, parent)
        self._kind = None
        if isinstance(modifier_invocation.modifier_name, SolcIdentifier):
            self._modifier_name = Identifier(
                init, modifier_invocation.modifier_name, self
            )
        elif isinstance(modifier_invocation.modifier_name, SolcIdentifierPath):
            self._modifier_name = IdentifierPath(
                init, modifier_invocation.modifier_name, self
            )

        if modifier_invocation.arguments is None:
            self._arguments = None
        else:
            self._arguments = [
                ExpressionAbc.from_ast(init, argument, self)
                for argument in modifier_invocation.arguments
            ]

        self._reference_resolver.register_post_process_callback(self._post_process)

    def _post_process(self, callback_params: CallbackParams):
        from ..declaration.contract_definition import ContractDefinition
        from ..declaration.modifier_definition import ModifierDefinition

        referenced_declaration = self.modifier_name.referenced_declaration
        if isinstance(referenced_declaration, ModifierDefinition):
            self._kind = ModifierInvocationKind.MODIFIER_INVOCATION
        elif isinstance(referenced_declaration, ContractDefinition):
            self._kind = ModifierInvocationKind.BASE_CONSTRUCTOR_SPECIFIER
        else:
            assert False, f"Unexpected declaration type: {referenced_declaration}"

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._modifier_name
        if self._arguments is not None:
            for argument in self._arguments:
                yield from argument

    @property
    def parent(self) -> FunctionDefinition:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def kind(self) -> ModifierInvocationKind:
        """
        Returns:
            Kind of the modifier invocation.
        """
        assert self._kind is not None
        return self._kind

    @property
    def modifier_name(self) -> Union[Identifier, IdentifierPath]:
        """
        The returned IR node holds a reference to the [ModifierDefinition][woke.ast.ir.declaration.modifier_definition.ModifierDefinition] declaration of the modifier being invoked in the case of the [ModifierInvocationKind.MODIFIER_INVOCATION][woke.ast.enums.ModifierInvocationKind.MODIFIER_INVOCATION] kind.
        In the case of the [ModifierInvocationKind.BASE_CONSTRUCTOR_SPECIFIER][woke.ast.enums.ModifierInvocationKind.BASE_CONSTRUCTOR_SPECIFIER] kind, the returned IR node holds a reference to the [ContractDefinition][woke.ast.ir.declaration.contract_definition.ContractDefinition] whose constructor is being invoked.
        !!! example
            `ERC20` and `initializer` in the following code:
            ```solidity
            constructor() ERC20("MyToken", "MTK") initializer {
                // ...
            }
            ```
        Returns:
            IR node representing the name of the modifier.
        """
        return self._modifier_name

    @property
    def arguments(self) -> Optional[List[ExpressionAbc]]:
        """
        !!! warning
            Is `None` when there are no curly braces after the modifier name.
            ```solidity
            constructor() initializer {}
            ```

            Is an empty list when there are curly braces but no arguments.
            ```solidity
            constructor() initializer() {}
            ```
        !!! example
            `:::solidity "MyToken"` and `:::solidity "MTK"` in the following code:
            ```solidity
            constructor() ERC20("MyToken", "MTK") initializer {
                // ...
            }
            ```
        Returns:
            Arguments of the base constructor or modifier invocation (if any).
        """
        return self._arguments



from __future__ import annotations

import re
from collections import deque
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Deque, Optional, Set, Tuple, Union

if TYPE_CHECKING:
    from .inheritance_specifier import InheritanceSpecifier
    from .modifier_invocation import ModifierInvocation
    from .override_specifier import OverrideSpecifier
    from .source_unit import SourceUnit
    from .using_for_directive import UsingForDirective
    from ..type_name.user_defined_type_name import UserDefinedTypeName

from intervaltree import IntervalTree

from woke.ast.ir.abc import SolidityAbc
from woke.ast.ir.declaration.abc import DeclarationAbc
from woke.ast.ir.reference_resolver import CallbackParams, ReferenceResolver
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import AstNodeId, SolcIdentifierPath

IDENTIFIER_RE = re.compile(r"[a-zA-Z$_][a-zA-Z0-9$_]*".encode("utf-8"))


class IdentifierPathPart:
    """
    A class representing a part of an identifier path. Is almost the same as [Identifier][woke.ast.ir.expression.identifier.Identifier], but it is not generated in the AST output of the compiler and so it is not an IR node.
    """

    _reference_resolver: ReferenceResolver
    _underlying_node: Union[IdentifierPath, UserDefinedTypeName]
    _referenced_declaration_id: Optional[AstNodeId]
    _cu_hash: bytes
    _file: Path
    _byte_location: Tuple[int, int]
    _name: str

    def __init__(
        self,
        underlying_node: Union[IdentifierPath, UserDefinedTypeName],
        byte_location: Tuple[int, int],
        name: str,
        referenced_declaration_id: AstNodeId,
        reference_resolver: ReferenceResolver,
        cu_hash: bytes,
        file: Path,
    ):
        self._underlying_node = underlying_node
        self._reference_resolver = reference_resolver
        self._referenced_declaration_id = referenced_declaration_id
        self._cu_hash = cu_hash
        self._file = file
        self._byte_location = byte_location
        self._name = name

        self._reference_resolver.register_post_process_callback(self._post_process)

    def _post_process(self, callback_params: CallbackParams):
        referenced_declaration = self.referenced_declaration
        if isinstance(referenced_declaration, DeclarationAbc):
            referenced_declaration.register_reference(self)
            self._reference_resolver.register_destroy_callback(
                self.file, partial(self._destroy, referenced_declaration)
            )

    def _destroy(self, referenced_declaration: DeclarationAbc) -> None:
        referenced_declaration.unregister_reference(self)

    @property
    def underlying_node(self) -> Union[IdentifierPath, UserDefinedTypeName]:
        """
        Returns:
            Underlying IR node (parent) of this identifier path part.
        """
        return self._underlying_node

    @property
    def file(self) -> Path:
        """
        The absolute path to the source file that contains the parent IR node of this identifier path part.
        Returns:
            Absolute path to the file containing this identifier path part.
        """
        return self._file

    @property
    def byte_location(self) -> Tuple[int, int]:
        """
        The start and end byte offsets of this identifier path part in the source file. `{node}.byte_location[0]` is the start byte offset, `{node}.byte_location[1]` is the end byte offset.

        `{node}.byte_location[1]` is always greater than or equal to `{node}.byte_location[0]`.
        Returns:
            Tuple of the start and end byte offsets of this node in the source file.
        """
        return self._byte_location

    @property
    def name(self) -> str:
        """
        !!! example
            `Contract` or `Event` in `Contract.Event`.
        Returns:
            Name of the identifier path part as it appears in the source code.
        """
        return self._name

    @property
    def referenced_declaration(self) -> Union[DeclarationAbc, SourceUnit]:
        """
        !!! example
            In the case of `Contract.Struct` [IdentifierPath][woke.ast.ir.meta.identifier_path.IdentifierPath], the referenced declaration of `Struct` is the declaration of the struct `Struct` in the contract `Contract`.
        !!! example
            Can be a [SourceUnit][woke.ast.ir.meta.source_unit.SourceUnit] in the following case:
            ```solidity
            import * as Utils from "./Utils.sol";
            ```
        Returns:
            Declaration referenced by this identifier path part.
        """
        from .source_unit import SourceUnit

        assert self._referenced_declaration_id is not None
        node = self._reference_resolver.resolve_node(
            self._referenced_declaration_id, self._cu_hash
        )
        assert isinstance(node, (DeclarationAbc, SourceUnit))
        return node


class IdentifierPath(SolidityAbc):
    """
    Identifier path represents a path name of identifiers separated by dots. It was introduced in Solidity 0.8.0 to replace
    [UserDefinedTypeName][woke.ast.ir.type_name.user_defined_type_name.UserDefinedTypeName] in some cases.
    """

    _ast_node: SolcIdentifierPath
    _parent: Union[
        InheritanceSpecifier,
        ModifierInvocation,
        OverrideSpecifier,
        UsingForDirective,
        UserDefinedTypeName,
    ]

    _name: str
    _referenced_declaration_id: AstNodeId
    _parts: IntervalTree

    def __init__(
        self,
        init: IrInitTuple,
        identifier_path: SolcIdentifierPath,
        parent: SolidityAbc,
    ):
        super().__init__(init, identifier_path, parent)
        self._name = identifier_path.name
        self._referenced_declaration_id = identifier_path.referenced_declaration
        assert self._referenced_declaration_id >= 0

        self._reference_resolver.register_post_process_callback(self._post_process)

    def _post_process(self, callback_params: CallbackParams):
        def find_referenced_source_unit(
            searched_name: str, start_source_unit: SourceUnit
        ) -> SourceUnit:
            source_units_queue: Deque[SourceUnit] = deque([start_source_unit])
            processed_source_units: Set[Path] = {start_source_unit.file}
            referenced_declaration = None

            while source_units_queue and referenced_declaration is None:
                source_unit = source_units_queue.popleft()

                for import_ in source_unit.imports:
                    if import_.unit_alias == searched_name:
                        referenced_declaration = callback_params.source_units[
                            import_.imported_file
                        ]
                        break
                    for symbol_alias in import_.symbol_aliases:
                        if symbol_alias.local == searched_name:
                            ref = symbol_alias.foreign.referenced_declaration
                            assert isinstance(ref, SourceUnit)
                            referenced_declaration = ref

                    if referenced_declaration is not None:
                        break

                    if import_.imported_file not in processed_source_units:
                        source_units_queue.append(
                            callback_params.source_units[import_.imported_file]
                        )
                        processed_source_units.add(import_.imported_file)

            assert referenced_declaration is not None
            return referenced_declaration

        from ..meta.source_unit import SourceUnit

        matches = list(IDENTIFIER_RE.finditer(self._source))
        groups_count = len(matches)
        assert groups_count > 0

        self._parts = IntervalTree()
        start_source_unit = callback_params.source_units[self._file]

        ref = self.referenced_declaration
        refs = []
        for _ in range(groups_count):
            refs.append(ref)
            if ref is not None:
                ref = ref.parent

        for match, ref in zip(matches, reversed(refs)):
            name = match.group(0).decode("utf-8")

            if ref is None:
                start_source_unit = find_referenced_source_unit(name, start_source_unit)
                referenced_node = start_source_unit
            elif isinstance(ref, (DeclarationAbc, SourceUnit)):
                referenced_node = ref
            else:
                raise TypeError(
                    f"Unexpected type of referenced declaration: {type(ref)}"
                )

            node_path_order = self._reference_resolver.get_node_path_order(
                AstNodeId(referenced_node.ast_node_id),
                referenced_node.cu_hash,
            )
            referenced_node_id = (
                self._reference_resolver.get_ast_id_from_cu_node_path_order(
                    node_path_order, self._cu_hash
                )
            )

            start = self.byte_location[0] + match.start()
            end = self.byte_location[0] + match.end()
            self._parts[start:end] = IdentifierPathPart(
                self,
                (start, end),
                name,
                referenced_node_id,
                self._reference_resolver,
                self._cu_hash,
                self._file,
            )

    @property
    def parent(
        self,
    ) -> Union[
        InheritanceSpecifier,
        ModifierInvocation,
        OverrideSpecifier,
        UsingForDirective,
        UserDefinedTypeName,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def name(self) -> str:
        """
        Returns:
            Name (as it appears in the source code) of the user-defined type referenced by this identifier path.
        """
        return self._name

    @property
    def identifier_path_parts(self) -> Tuple[IdentifierPathPart, ...]:
        """
        Returns:
            Parts of the identifier path.
        """
        return tuple(interval.data for interval in sorted(self._parts))

    def identifier_path_part_at(self, byte_offset: int) -> Optional[IdentifierPathPart]:
        """
        Parameters:
            byte_offset: Byte offset in the source code.
        Returns:
            Identifier path part at the given byte offset, if any.
        """
        intervals = self._parts.at(byte_offset)
        assert len(intervals) <= 1
        if len(intervals) == 0:
            return None
        return intervals.pop().data

    @property
    def referenced_declaration(self) -> DeclarationAbc:
        """
        Returns:
            Declaration referenced by this identifier path.
        """
        node = self._reference_resolver.resolve_node(
            self._referenced_declaration_id, self._cu_hash
        )
        assert isinstance(node, DeclarationAbc)
        return node



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Iterator, Set, Tuple, Union

from woke.ast.enums import ModifiesStateFlag
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.expression.abc import ExpressionAbc
from woke.ast.ir.statement.abc import StatementAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import SolcWhileStatement

if TYPE_CHECKING:
    from .block import Block
    from .do_while_statement import DoWhileStatement
    from .for_statement import ForStatement
    from .if_statement import IfStatement
    from .unchecked_block import UncheckedBlock


class WhileStatement(StatementAbc):
    """
    !!! example
        Lines 2-3 in the following code:
        ```solidity linenums="1"
        function foo(uint x) public pure {
            while (x % 2 == 0)
                x /= 2;
        }
        ```
    """

    _ast_node: SolcWhileStatement
    _parent: Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]

    _body: StatementAbc
    _condition: ExpressionAbc

    def __init__(
        self,
        init: IrInitTuple,
        while_statement: SolcWhileStatement,
        parent: SolidityAbc,
    ):
        super().__init__(init, while_statement, parent)
        self._body = StatementAbc.from_ast(init, while_statement.body, self)
        self._condition = ExpressionAbc.from_ast(init, while_statement.condition, self)

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._body
        yield from self._condition

    @property
    def parent(
        self,
    ) -> Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def body(self) -> StatementAbc:
        """
        Returns:
            Body of the while statement.
        """
        return self._body

    @property
    def condition(self) -> ExpressionAbc:
        """
        Returns:
            Condition of the while statement.
        """
        return self._condition

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return self.body.modifies_state | self.condition.modifies_state

    def statements_iter(self) -> Iterator["StatementAbc"]:
        yield self
        yield from self._body.statements_iter()



import logging
from typing import List, Set

from woke.analysis.detectors import DetectorAbc, DetectorResult, detector
from woke.analysis.detectors.ownable import statement_is_publicly_executable
from woke.ast.enums import GlobalSymbolsEnum
from woke.ast.ir.expression.function_call import FunctionCall
from woke.ast.ir.statement.abc import StatementAbc

logger = logging.getLogger(__name__)


@detector(-100, "unsafe-selfdestruct")
class UnsafeSelfdestructDetector(DetectorAbc):
    """
    Selfdestruct call is not protected.
    """

    _detections: Set[DetectorResult]

    def __init__(self):
        self._detections = set()

    def report(self) -> List[DetectorResult]:
        return list(self._detections)

    def visit_function_call(self, node: FunctionCall):
        if node.function_called not in {
            GlobalSymbolsEnum.SELFDESTRUCT,
            GlobalSymbolsEnum.SUICIDE,
        }:
            return

        current_node = node
        while current_node is not None:
            if isinstance(current_node, StatementAbc):
                break
            current_node = current_node.parent
        if current_node is None:
            return
        if not statement_is_publicly_executable(current_node):
            return

        self._detections.add(DetectorResult(node, "Selfdestruct call is not protected"))
